#include "page_heap.h"
#include "shm_helper.h"
#include "size_map.h"
#include "shm_config.h"
#include "shm.h"
#include <inttypes.h>
namespace bg
{
    namespace detail
    {
        PageHeap::PageHeap()
        {
        }

        bg::detail::Span* PageHeap::GetSpanMap(size_t start_page)
        {
            SHM_ASSERT_RETURN_NULL(start_page <= ShmConfig::MAX_PAGE_COUNT);

            auto node1 = m_span_map.lv0[PTR_TO_LV0(start_page)];
            if(node1 == nullptr)
            {
                return nullptr;
            }

            auto node2 = node1->lv1[PTR_TO_LV1(start_page)];
            if(node2 == nullptr)
            {
                return nullptr;
            }

            auto node3 = node2->lv2[PTR_TO_LV2(start_page)];
            return node3;
        }

        bool PageHeap::SetSpanMap(size_t page, bg::detail::Span* span)
        {
            SHM_ASSERT_RETURN_FALSE(page <= ShmConfig::MAX_PAGE_COUNT);

            m_span_map.lv0[PTR_TO_LV0(page)]->lv1[PTR_TO_LV1(page)]->lv2[PTR_TO_LV2(page)] = span;
            return true;
        }

        bg::detail::Span* PageHeap::AllocateSpan(size_t page_count)
        {
            auto span = GetListSpan(page_count);
            if(!span)
            {
                if(!GrowHeap(page_count) || !(span = GetListSpan(page_count)))
                {
                    SHM_ERROR("cannot allocate span, page count: %lu.", page_count);
                    return nullptr;
                }

            }
            span->RemoveSelfForList();

            do
            {
                int64_t diff = span->GetPageCount() - page_count;
                if(diff == 0)
                {
                    break;
                }

                Span* free_list = m_span_allocator.Allocate();
                if(!free_list)
                {
                    break;
                }

                span->SetPageCount(page_count);
                if(page_count)
                {
                    SetSpanMap(span->GetLastPage(), span);
                }

                size_t start_page = span->GetLastPage() + 1;
                new (free_list) Span(start_page, diff);
                SetSpanMap(start_page, free_list);

                Span* list = &m_large_list;
                SetSpanMap(free_list->GetLastPage(), free_list);
                if(diff <= 1 || diff < MAX_FREE_LIST_COUNT)
                {
                    list = &m_free_lists[diff];
                }
                list->InstertNextSpan(free_list);

            } while(false);

            span->SetUseState(true);
            return span;
        }


        void PageHeap::DeallocateSpan(bg::detail::Span* s)
        {
            s->SetUseState(false);
            bg::detail::Span* pre_span = GetSpanMap(s->GetStartPage() - 1);
            if(pre_span && !pre_span->GetUseState())
            {
                s->LeftMovePages(pre_span->GetPageCount());

                pre_span->RemoveSelfForList();
                m_span_allocator.Destroy(pre_span);

                SetSpanMap(s->GetStartPage(), s);
            }

            bg::detail::Span* next_span = GetSpanMap(s->GetLastPage() + 1);
            if(next_span && !next_span->GetUseState())
            {
                s->RightMovePages(next_span->GetPageCount());

                next_span->RemoveSelfForList();

                m_span_allocator.Destroy(next_span);

                SetSpanMap(s->GetLastPage(), s);
            }

            bg::detail::Span* p_list;
            p_list = &m_large_list;
            if(s->GetPageCount() < MAX_FREE_LIST_COUNT)
            {
                p_list = &m_free_lists[s->GetPageCount()];
            }
            p_list->InstertNextSpan(s);

        }

        void PageHeap::RegisterSpan(bg::detail::Span* span)
        {
            size_t start = span->GetStartPage() + 1;
            size_t end = span->GetLastPage();
            while(end > start)
            {
                size_t index = start;
                if(!SetSpanMap(index, span))
                {
                    break;
                }
                ++start;
            }
        }

        bool PageHeap::GrowHeap(size_t page_count)
        {
            size_t real_bytes{};
            size_t page_start{};
            void* rawmemoey{};
            if(page_count > ShmConfig::MAX_PAGE_COUNT)
            {
                return false;
            }

            if(page_count >= MAX_FREE_LIST_COUNT)
            {
                real_bytes = PAGES_TO_BYTES(page_count);
                rawmemoey = bg::detail::ShmAllocateRawMemory(&real_bytes, PAGE_BYTES);
                if(!rawmemoey)
                {
                    return false;
                }
            }
            else
            {
                real_bytes = DEFULT_ALLOC_BYTES;
                rawmemoey = bg::detail::ShmAllocateRawMemory(&real_bytes, PAGE_BYTES);
                if(!rawmemoey)
                {
                    real_bytes = PAGES_TO_BYTES(page_count);
                    rawmemoey = bg::detail::ShmAllocateRawMemory(&real_bytes, PAGE_BYTES);
                    if(!rawmemoey)
                    {
                        return false;
                    }
                }
            }


            if((int64_t)rawmemoey >> 48)
            {
                SHM_ERROR("cannot allocate enough space in span map, memory(%p, %#" PRIx64") leaked.", rawmemoey, real_bytes);
                return false;
            }

            page_start = BYTES_TO_PAGES((int64_t)rawmemoey);
            size_t real_pages = BYTES_TO_PAGES(real_bytes);
            size_t page_end = page_start + real_pages;

            size_t index = page_start;
            for(; index < page_end; ++index)
            {
                size_t lv0_index = PTR_TO_LV0(index);
                size_t lv1_index = PTR_TO_LV1(index);


                if(m_span_map.lv0[lv0_index] == nullptr)
                {
                    size_t node1_size = sizeof(RadixTree::NodeV1);
                    RadixTree::NodeV1* p_node1 = (RadixTree::NodeV1*)bg::detail::ShmAllocateRawMemory(
                        &node1_size, sizeof(RadixTree::NodeV1));

                    if(!p_node1)
                    {
                        break;
                    }
                    m_span_map.lv0[lv0_index] = p_node1;
                    new (p_node1) RadixTree::NodeV1();

                }
                if(m_span_map.lv0[lv0_index]->lv1[lv1_index] == nullptr)
                {
                    size_t node2_size = sizeof(RadixTree::NodeV2);
                    RadixTree::NodeV2* p_node2 = (RadixTree::NodeV2*)bg::detail::ShmAllocateRawMemory(&node2_size, sizeof(RadixTree::NodeV2));
                    if(!p_node2)
                    {
                        break;
                    }
                    m_span_map.lv0[lv0_index]->lv1[lv1_index] = p_node2;
                    new (p_node2) RadixTree::NodeV2();

                }
            }

            if(index < page_end)
            {
                SHM_ERROR("cannot allocate new span, memory(%p, %#" PRIx64") leaked.", rawmemoey, real_bytes);
                return false;
            }

            bg::detail::Span* new_span = m_span_allocator.Allocate();
            if(!new_span)
            {
                SHM_ERROR("cannot allocate new span, memory(%p, %#" PRIx64") leaked.", rawmemoey, real_bytes);
                return false;
            }
            new (new_span) Span(page_start, real_pages);

            SetSpanMap(page_start, new_span);

            if(real_pages > 1)
            {
                SetSpanMap(page_end - 1, new_span);
            }

            this->DeallocateSpan(new_span);
            return true;
        }



        bg::detail::Span* PageHeap::GetSpanByFreeList(size_t page_count)
        {
            if(page_count < MAX_FREE_LIST_COUNT)
            {
                Span* free_span = &m_free_lists[page_count];

                if(free_span->GetNextSpan() == free_span)
                {
                    for(free_span = &m_free_lists[page_count + 1]; free_span != &m_large_list; ++free_span)
                    {
                        if(free_span != free_span->GetNextSpan())
                        {
                            return free_span->GetNextSpan();
                        }
                    }
                }
                else
                {
                    return free_span->GetNextSpan();
                }
            }
            return nullptr;
        }

        bg::detail::Span* PageHeap::GetSpanByLargeList(size_t page_count)
        {
            Span* result_span = nullptr;
            if(m_large_list.GetNextSpan() != &m_large_list)
            {
                for(Span* free_span = m_large_list.GetNextSpan(); &m_large_list != free_span; free_span = free_span->GetNextSpan())
                {
                    if(page_count > free_span->GetPageCount())
                    {
                        continue;
                    }
                    if(!result_span || free_span->GetPageCount() < result_span->GetPageCount() ||
                       (result_span->GetPageCount() == free_span->GetPageCount() && free_span->GetStartPage() < result_span->GetStartPage()))
                    {
                        result_span = free_span;
                    }

                }

            }
            return result_span;
        }


    }


}