#include "shm_manager.h"
#include "shm_object.h"
#include "shm_config.h"
#include <inttypes.h>

extern bg::detail::ShmContext g_shm_ctx;
namespace bg
{
    namespace detail
    {

        ShmManager::ShmManager(const bg::ShmOptions& options) : m_options(options)
        {
            snprintf(this->m_magic, 0x40uLL, "ByteGame/ShmManager-%s", this->m_options.identifier);
        }

        ShmManager::~ShmManager()
        {
            if(m_singletons)
            {
                m_singletons->clear();
                bg::ShmFree(m_singletons);
                this->m_singletons = nullptr;
            }
            for(size_t i = 0LL; this->m_block_count > i; ++i)
            {
                DeleteBlock(i);
            }
        }

        void* ShmManager::AllocateRawMemory(size_t* bytes, size_t alignment)
        {
            void* result = nullptr;
            if((alignment & (alignment - 1)) != 0)
            {
                SHM_ERROR("invalid alignment(%#" PRIx64").", alignment);
                return result;
            }
            *bytes = PALIGN_UP(*bytes, alignment);

            if(m_block_count)
            {
                result = AllocateInBlock(m_block_count - 1, *bytes, alignment);
                if(result)
                {
                    return result;
                }

            }
            if(CreateBlock(m_block_count, *bytes))
            {
                size_t index = m_block_count > 1 ? m_block_count - 1 : 0;
                result = AllocateInBlock(index, *bytes, alignment);
                if(!result)
                {
                    SHM_ERROR("newly created block(%lu) still cannot satisfy requirement, addr(%#" PRIx64"), used size(%#" PRIx64"), real size(%#" PRIx64"), real bytes(%#" PRIx64"), alignment(%#" PRIx64").",
                    m_block_count, m_blocks[index].addr, m_blocks[index].used_size, m_blocks[index].real_size);
                    return result;
                }

            }
            else
            {
                SHM_ERROR("failed to create block(%lu), real bytes(%#" PRIx64").", this->m_block_count, *bytes);
            }
            return result;
        }

        bool ShmManager::HasSingleton(const bg::detail::TypeName& type)
        {
            auto itor = m_singletons->find(type);
            if(itor == m_singletons->end())
            {
                return false;
            }
            return true;
        }

        void* ShmManager::GetSingleton(const bg::detail::TypeName& type, size_t bytes, bool* first_call)
        {
            if(!m_singletons)
            {
                m_singletons = (bg::detail::ShmManager::SingletonMap*)bg::ShmMalloc(sizeof(bg::detail::ShmManager::SingletonMap));
                if(!m_singletons)
                {
                    m_singletons = nullptr;
                    return nullptr;
                }

                new(m_singletons) SingletonMap();
            }

            auto itor = m_singletons->find(type);
            if(itor != m_singletons->end())
            {
                return itor->second;
            }

            void* memory = ShmMalloc(bytes);
            if(!memory)
            {
                return nullptr;
            }

            m_singletons->operator[](type) = memory;
            if(first_call)
            {
                *first_call = 1;
                SHM_INFO("singleton(%s: %p) created.", type.c_str(), memory);
            }
            return memory;
        }

        void ShmManager::FreeSingleton(const bg::detail::TypeName& type)
        {
            if(m_singletons)
            {
                auto itor = m_singletons->find(type);
                if(itor == m_singletons->end())
                {
                    return;
                }
                SHM_INFO("singleton(%s: %p) destroyed.", type.c_str(), itor->second);

                Free(itor->second);
                m_singletons->erase(itor);

            }

            return;

        }

        bool ShmManager::OnResume(const char* identifier)
        {
            char magic[64]; // [rsp+0h] [rbp-60h] BYREF

            int size = snprintf(magic, 0x40uLL, "ByteGame/ShmManager-%s", identifier);
            if(!strncmp(magic, this->m_magic, size))
            {
                size_t index = 0;
                for(; index < m_block_count; ++index)
                {
                    if(!AttachBlock(index))
                    {
                        SHM_ERROR("failed to attach block(%lu), address(%#" PRIx64"), used size(%#" PRIx64"), real size(%#" PRIx64"), mmap size(%#" PRIx64").", index, m_blocks[index].addr, m_blocks[index].used_size, m_blocks[index].real_size, m_blocks[index].mmap_size);
                        return false;
                    }

                }
                if(m_block_count <= index)
                {
                    m_version++;
                    return true;
                }
            }

            SHM_ERROR("magic mismatch, expected: %s, actual: %s.", magic, this->m_magic);
            return false;
        }

        bool ShmManager::OnCreate(void)
        {
            return m_size_map.Init();
        }

        void* ShmManager::Malloc(size_t bytes)
        {
            if(bytes > CLASS_MAX_BYTES)
            {
                bg::detail::Span* span = m_page_heap.AllocateSpan(BYTES_TO_PAGES(bytes + PAGE_BYTES));
                if(span)
                {
                    return (void*)PAGES_TO_BYTES(span->GetStartPage());
                }
                return nullptr;
            }

            size_t index = m_size_map.GetClassByBytes(bytes);

            SHM_ASSERT_RETURN_NULL(index < CLASS_MAX_COUNT);
            return AllocateChunk(index);
        }

        void ShmManager::Free(void* ptr)
        {
            bg::detail::Span* span = m_page_heap.GetSpanMap(BYTES_TO_PAGES((uintptr_t)ptr));
            SHM_ASSERT_RETURN_VOID(span);
            if(span->GetSizeClass() == MAX_SHM_BLOCK_COUNT)
            {
                m_page_heap.DeallocateSpan(span);
            }
            else
            {
                DeallocateChunk(span, ptr);
            }

        }

        void* ShmManager::Realloc(void* old_ptr, size_t new_bytes)
        {
            if(!old_ptr)
            {
                return bg::ShmMalloc(new_bytes);
            }
            if(!new_bytes)
            {
                bg::ShmFree((void*)old_ptr);
                return nullptr;
            }

            bg::detail::Span* span = m_page_heap.GetSpanMap(BYTES_TO_PAGES((uintptr_t)old_ptr));
            SHM_ASSERT_RETURN_NULL(span);
            size_t old_size = 0;
            if(span->GetSizeClass() == MAX_SHM_BLOCK_COUNT)
            {
                old_size = PAGES_TO_BYTES(span->GetPageCount());
            }
            else
            {
                old_size = m_size_map.ClassToSize(span->GetSizeClass());
            }

            if(new_bytes >= (old_size >> 1) && new_bytes <= old_size)
            {
                return old_ptr;
            }

            void* new_ptr = g_shm_ctx.mgr->Malloc(new_bytes);
            if(!new_ptr)
            {
                return nullptr;
            }

            memcpy(new_ptr, (const void*)old_ptr, old_size);
            g_shm_ctx.mgr->Free(old_ptr);
            return new_ptr;
        }

        void* ShmManager::Calloc(size_t n, size_t bytes)
        {
            size_t toal_bytes = bytes * n;
            if(bytes && n != toal_bytes / bytes)
            {
                return nullptr;
            }
            void* memory = Malloc(toal_bytes);
            if(memory)
            {
                memset(memory, 0, toal_bytes);
            }

            return memory;
        }

        bool ShmManager::DeleteBlock(size_t index)
        {
            SHM_ASSERT_RETURN_FALSE(index < m_block_count);
            char path[256];
            if(m_blocks[index].addr)
            {
                m_options.AddFixedAddress((uintptr_t)m_blocks[index].addr);
                snprintf(path, 0x100uLL, "%s-%03zu.mmap", this->m_options.identifier, index);
                bg::detail::ShmObjectDelete(m_blocks[index].addr, m_blocks[index].mmap_size, path);
                SHM_INFO("block(%lu) deleted, address(%#" PRIx64"), used size(%#" PRIx64"), real size(%#" PRIx64"), mmap size(%#" PRIx64"), path(%s).",
                         index, m_blocks[index].addr, m_blocks[index].used_size, m_blocks[index].real_size, m_blocks[index].mmap_size, path);
                m_blocks[index].addr = nullptr;
            }


            return true;
        }

        void* ShmManager::AllocateInBlock(size_t index, size_t bytes, size_t alignment)
        {
            SHM_ASSERT_RETURN_NULL(index < m_block_count);
            uintptr_t addr_ptr = (uintptr_t)m_blocks[index].addr;
            uintptr_t ptr = addr_ptr + m_blocks[index].used_size;
            ptr = PALIGN_UP(ptr, alignment);
            uintptr_t new_ptr = ptr + bytes;

            if(new_ptr > m_blocks[index].mmap_size + addr_ptr)
            {
                return nullptr;
            }
            do
            {
                if(m_blocks[index].real_size + addr_ptr >= new_ptr)
                {
                    break;
                }
                size_t new_size = m_options.shm_block_grow_size + m_blocks[index].real_size;
                if(new_ptr > new_size + addr_ptr)
                {
                    new_size = new_ptr - addr_ptr;
                    new_size = PALIGN_UP(new_size, m_options.shm_block_grow_size);
                }
                new_size = std::min(new_size, m_blocks[index].mmap_size);

                if(ResizeBlock(index, new_size))
                {
                    break;
                }
                SHM_ERROR("failed to resize block(%lu), used size(%#" PRIx64"), real size(%#" PRIx64"), mmap size(%#" PRIx64"), new size(%#" PRIx64").",
                          index, m_blocks[index].used_size, m_blocks[index].real_size, m_blocks[index].mmap_size, new_size);
                return nullptr;

            } while(false);

            m_blocks[index].used_size = new_ptr - addr_ptr;
            return (void*)ptr;
        }

        bool ShmManager::CreateBlock(size_t index, size_t min_size)
        {
            char path[256];
            SHM_ASSERT_RETURN_FALSE(index < MAX_SHM_BLOCK_COUNT);

            size_t real_size = std::max(min_size, m_options.shm_block_grow_size);
            size_t mmap_size = std::max(min_size, m_options.shm_block_mmap_size);
            uintptr_t fixed_addr = 0ll;
            if(!m_options.PopFixedAddress(&fixed_addr))
            {
                SHM_WARN("fixed addresses have been used up.");
            }

            snprintf(path, 0x100uLL, "%s-%03zu.mmap", this->m_options.identifier, index);
            void* memory = bg::detail::ShmObjectCreate(path, fixed_addr, &real_size, &mmap_size);
            if(!memory)
            {

                if(fixed_addr)
                {
                    m_options.AddFixedAddress(fixed_addr);
                }
                return false;
            }

            m_blocks[index].mmap_size = mmap_size;
            m_blocks[index].real_size = real_size;
            m_blocks[index].used_size = 0;
            m_blocks[index].addr = memory;
            SHM_INFO("block(%" SCNu64") created, address(%#" PRIx64"), real size(%#" PRIx64"), mmap size(%#" PRIx64").", index, memory, real_size, mmap_size);

            if(this->m_block_count != index)
            {
                return true;
            }

            this->m_block_count = index + 1;

            return true;
        }

        bool ShmManager::AttachBlock(size_t index)
        {
            char path[256];
            SHM_ASSERT_RETURN_FALSE(index < m_block_count);
            if(m_blocks[index].addr)
            {
                size_t real_size = m_blocks[index].real_size;
                size_t mmap_size = m_blocks[index].mmap_size;
                snprintf(path, 0x100uLL, "%s-%03zu.mmap", this->m_options.identifier, index);
                void* memory = bg::detail::ShmObjectAttach(path, (uintptr_t)m_blocks[index].addr, &real_size, &mmap_size);
                return memory != nullptr;
            }

            return true;
        }

        bool ShmManager::ResizeBlock(size_t index, size_t new_size)
        {
            char path[256]; // [rsp+10h] [rbp-120h] BYREF
            SHM_ASSERT_RETURN_FALSE(index < m_block_count);

            if(m_blocks[index].mmap_size < new_size)
            {
                SHM_ERROR("new size(%#" PRIx64") too large, block(%lu), mmap size(%#" PRIx64").", new_size, index, m_blocks[index].mmap_size);
                return false;
            }
            else
            {
                snprintf(path, 0x100uLL, "%s-%03zu.mmap", this->m_options.identifier, index);
                if(bg::detail::ShmObjectResize(path, &new_size))
                {
                    SHM_INFO("block(%lu) resized, old size(%#" PRIx64"), new size(%#" PRIx64"), used size(%#" PRIx64"), mmap size(%#" PRIx64").",
                             index, m_blocks[index].addr, m_blocks[index].real_size, new_size, m_blocks[index].used_size, m_blocks[index].mmap_size);
                    this->m_blocks[index].real_size = new_size;
                    return true;
                }

            }
            return false;
        }

        void* ShmManager::AllocateChunk(size_t index)
        {
            ChunkCache::ClassCache* cache = &m_chunk_cache.m_caches[index];
            if(cache == (bg::detail::ChunkCache::ClassCache*)cache->free_list.GetNextSpan())
            {
                ChunkCache::ClassCache* cheche = &this->m_chunk_cache.m_caches[index];
                size_t size = this->m_size_map.ClassToSize(index);

                bg::detail::Span* span = m_page_heap.AllocateSpan(m_size_map.ClassToPages(index));
                if(span)
                {
                    m_page_heap.RegisterSpan(span);
                    span->SetSizeClass(static_cast<uint8_t>(index));

                    span->SetChunkList((void*)PAGES_TO_BYTES(span->GetStartPage()));

                    uintptr_t end_memory = (uintptr_t)PAGES_TO_BYTES(span->GetLastPage());

                    uintptr_t end_ptr = (uintptr_t)span->GetChunkList() + size;

                    uintptr_t start_ptr = (uintptr_t)span->GetChunkList();

                    uintptr_t ptr = start_ptr;

                    for(; end_ptr <= end_memory; )
                    {
                        ptr = start_ptr;
                        *((uintptr_t*)start_ptr) = end_ptr;
                        auto temp = end_ptr + size;
                        start_ptr = end_ptr;
                        end_ptr += size;
                    }

                    *(uintptr_t*)ptr = 0ll;
                    cheche->free_list.InstertNextSpan(span);
                    ++cheche->span_count;
                }
            }

            if(cache == (bg::detail::ChunkCache::ClassCache*)cache->free_list.GetNextSpan())
            {
                return nullptr;
            }

            auto free_span = cache->free_list.GetNextSpan();
            free_span->IncUsedCount();

            auto chunk_span = free_span->GetChunkList();
            free_span->SetChunkList(*(void**)chunk_span);
            if(free_span->GetChunkList() == nullptr)
            {
                cache->free_list.RemoveNextSpan();
                --this->m_chunk_cache.m_caches[index].span_count;
            }
            return chunk_span;
        }

        void ShmManager::DeallocateChunk(bg::detail::Span* span, void* ptr)
        {
            SHM_ASSERT_RETURN_VOID(span->GetUsedCount() > 0 && span->GetSizeClass() < CLASS_MAX_COUNT);

            span->DecUsedCout();
            auto chunk_list = span->GetChunkList();
            *(uintptr_t*)ptr = (uintptr_t)chunk_list;
            span->SetChunkList(ptr);

            if(chunk_list)
            {
                if(span->GetUsedCount() == 0)
                {
                    if(m_chunk_cache.m_caches[span->GetSizeClass()].span_count > 1)
                    {
                        span->RemoveSelfForList();
                        --m_chunk_cache.m_caches[span->GetSizeClass()].span_count;
                        span->SetSizeClass(MAX_SHM_BLOCK_COUNT);
                        span->SetChunkList(nullptr);
                        m_page_heap.DeallocateSpan(span);
                        return;
                    }
                }
            }
            else
            {
                if(span->GetUsedCount() == 0)
                {
                    if(m_chunk_cache.m_caches[span->GetSizeClass()].span_count)
                    {
                        span->SetSizeClass(MAX_SHM_BLOCK_COUNT);
                        span->SetChunkList(nullptr);
                        m_page_heap.DeallocateSpan(span);
                        return;
                    }

                }
                m_chunk_cache.m_caches[span->GetSizeClass()].free_list.GetNextSpan()->InstertNextSpan(span);
                m_chunk_cache.m_caches[span->GetSizeClass()].span_count++;
            }

        }


    }

    void* ShmMalloc(size_t bytes)
    {
        if(!g_shm_ctx.mgr)
        {
            SHM_ERROR("shm management system uninitialized.");
            return nullptr;
        }

        return g_shm_ctx.mgr->Malloc(bytes);
    }

    void ShmFree(void* ptr)
    {
        if(!g_shm_ctx.mgr)
        {
            SHM_ERROR("shm management system uninitialized.");
            return;
        }
        g_shm_ctx.mgr->Free(ptr);
        return;
    }

    void* ShmRealloc(void* old_ptr, size_t new_bytes)
    {
        if(!g_shm_ctx.mgr)
        {
            SHM_ERROR("shm management system uninitialized.");
            return nullptr;
        }
        return g_shm_ctx.mgr->Realloc(old_ptr, new_bytes);
    }

    void* ShmCalloc(size_t n, size_t bytes)
    {
        if(!g_shm_ctx.mgr)
        {
            SHM_ERROR("shm management system uninitialized.");
            return nullptr;
        }
        return g_shm_ctx.mgr->Calloc(n, bytes);
    }

}



