#include "vptr_manager.h"
#include "shm.h"
#include <memory>
#include <inttypes.h>
#define  VPTR_MAGIC 20230430
namespace bg
{
    namespace detail
    {
        VptrManager* mgr;
        class HighOrderContainer;
        std::unique_ptr<HighOrderContainer> hoc;

        class HighOrderContainer
        {
        public:
            VptrObjectContainerBase* FindContainer(const bg::detail::TypeName& type_name)
            {
                auto itor = m_type_to_containers.find(type_name);
                if(itor == m_type_to_containers.end())
                {
                    return itor->second->ShmNew();
                }
                SHM_INFO("type(%s) not found!", type_name.c_str());
                return nullptr;
            }


            bool FixContainerVptr(VptrObjectContainerBase* fix_ptr)
            {
                const auto& type_name = fix_ptr->GetTypeName();
                auto itor = m_type_to_containers.find(type_name);
                if(itor == m_type_to_containers.end())
                {
                    SHM_ERROR("type(%s) not found!", type_name.c_str());
                    return false;
                }

                const uintptr_t* new_vptr = reinterpret_cast<const uintptr_t*>(reinterpret_cast<const char*>(itor->second));
                uintptr_t* old_vptr = reinterpret_cast<uintptr_t*>(reinterpret_cast<char*>(fix_ptr));
                if(*old_vptr != *new_vptr)
                {
                    *old_vptr = *new_vptr;
                }
                SHM_INFO("type(%s), old vptr(%#" PRIx64"), new vptr(%#" PRIx64").", type_name.c_str(), old_vptr, new_vptr);
                return true;
            }

            VptrObjectContainerBase* NewShmContainer(const bg::detail::TypeName& type_name)
            {
                auto itor = m_type_to_containers.find(type_name);

                SHM_ASSERT_RETURN_NULL(itor != m_type_to_containers.end());

                return itor->second;
            }

            void RegisterContainer(VptrObjectContainerBase* obj)
            {
                auto& type_name = obj->GetTypeName();
                auto itor = m_type_to_containers.find(type_name);
                SHM_ASSERT_RETURN_VOID(itor != m_type_to_containers.end());

                m_type_to_containers[type_name] = obj;

            }

            template<typename Key, typename Value, typename Hash = std::hash<Key>, typename Pred = std::equal_to<Key>>
            using Container = std::unordered_map<Key, Value, Hash, Pred, ShmAllocator<std::pair<const Key, Value>>>;
            Container<TypeName, VptrObjectContainerBase*, TypeNameHasher> m_type_to_containers;
        };


        VptrObjectContainerBase::VptrObjectContainerBase(const TypeName& type, bool in_shm) :m_version(0LL), m_type(type)
        {
            if(in_shm)
            {
                std::vector<void*, bg::detail::ShmAllocator<void*> >* vector = (std::vector<void*, bg::detail::ShmAllocator<void*> > *)bg::ShmMalloc(sizeof(std::vector<void*, bg::detail::ShmAllocator<void*> >));
                if(vector)
                {
                    new (vector) std::vector<void*, bg::detail::ShmAllocator<void*> >();
                }
                m_objects = vector;
                this->m_version = bg::detail::ShmCurrentVersion();
            }
            else
            {
                if(!hoc)
                {
                    hoc = std::make_unique<HighOrderContainer>();
                }
                hoc->RegisterContainer(this);
            }
        }


        VptrObjectContainerBase::~VptrObjectContainerBase()
        {
            if(m_objects)
            {
                if(!m_objects->empty())
                {
                    SHM_ERROR("there are still %lu objects in container(%s)!", m_objects->size(), m_type.c_str());
                }
                m_objects->clear();
                bg::ShmFree(m_objects);
            }
            if(m_base_types)
            {
                m_base_types->clear();
                delete m_base_types;
            }
        }

        void VptrObjectContainerBase::FixVptr()
        {
            if(m_version == bg::detail::ShmCurrentVersion())
            {
                SHM_DEBUG("container(%p: %s) already fixed, skip.", this, m_type.c_str());
            }
            else if(FixSelf())
            {
                FixObjects();
                this->m_version = bg::detail::ShmCurrentVersion();
            }
            else
            {
                SHM_INFO("container(%p: %s) seems to be defined in a dlopen() so, try fixing after dlopen() called.", this, m_type.c_str());
            }

        }

        const BaseTypeInfo* VptrObjectContainerBase::GetBaseTypeInfo(const TypeName& base_type)
        {
            auto container = hoc->FindContainer(base_type);
            if(!container)
            {
                return nullptr;
            }
            m_base_types = container->m_base_types;

            if(m_base_types->empty())
            {
                return nullptr;
            }
            for(const auto& type : *m_base_types)
            {
                if(type.type == base_type)
                {
                    return &type;
                }
            }
            return nullptr;

        }

        void VptrObjectContainerBase::RecordObject(void* ptr)
        {
            if(ptr)
            {
                m_objects->push_back(ptr);
                size_t size = m_objects->size();
                *(reinterpret_cast<uintptr_t*>(ptr) - 1) = size;
                *(reinterpret_cast<uintptr_t*>(ptr) - 2) = VPTR_MAGIC;
                SHM_INFO("object(%p) added, type: %s, index: %lu.", ptr, m_type.c_str(), size);
            }
        }

        void VptrObjectContainerBase::RemoveObject(void* ptr)
        {
            if(ptr)
            {
                uintptr_t maric = *((uintptr_t*)ptr - 2);
                if(maric == VPTR_MAGIC)
                {
                    size_t index = *((uintptr_t*)ptr - 1);
                    size_t size = m_objects->size();
                    if(index < size)
                    {
                        if((*m_objects)[index] == ptr)
                        {
                            bg::detail::VptrObjectContainerBase::DoRemoveObject(index);
                            return;
                        }
                        SHM_WARN("address mismatch(%p <=> %p), type: %s, index: %#x.", (*m_objects)[index], ptr, m_type.c_str(), index);
                    }
                    else
                    {
                        SHM_ERROR("wrong index(%#x), size: %#" PRIx64", type: %s.", index, size, m_type.c_str());
                    }
                    for(int i = 0; i < size; ++i)
                    {
                        if((*m_objects)[i] == ptr)
                        {
                            SHM_INFO("found correct index: %#" PRIx64".", i);
                            bg::detail::VptrObjectContainerBase::DoRemoveObject(i);
                        }
                    }
                    SHM_ERROR("no stored object found with address(%p).", ptr);
                    return;
                }
                SHM_ERROR("invalid magic(%#x) for object(%p).", maric, ptr);
                return;
            }
        }

        bool VptrObjectContainerBase::FixSelf()
        {
            SHM_INFO(">>>>> fixing vptr of container(%p: %s)...", this, m_type.c_str());
            const char* result = "";
            if(bg::detail::hoc && bg::detail::hoc->FixContainerVptr(this))
            {
                SHM_INFO("<<<<< vptr of container(%p: %s) fixed.", this, m_type.c_str());
                return true;
            }
            SHM_ERROR("<<<<< vptr of container(%p: %s) not fixed.", this, m_type.c_str());
            return false;
        }

        void VptrObjectContainerBase::DoRemoveObject(size_t index)
        {
            size_t size = m_objects->size();
            SHM_ASSERT_RETURN_VOID(index < m_objects->size());
            if(index == size - 1)
            {
                m_objects->pop_back();
            }
            else
            {
                auto ptr = (*m_objects)[size - 1];
                *((uintptr_t*)ptr - 1) = index;
                (*m_objects)[index] = ptr;
                m_objects->pop_back();
            }
        }



        bool VptrManager::Init()
        {
            if(!bg::detail::mgr)
            {
                bg::detail::mgr = (bg::detail::VptrManager*)bg::ShmGetSingleton<bg::detail::VptrManager>();
            }
            return bg::detail::mgr != nullptr;
        }

        void VptrManager::Fini()
        {
            if(bg::detail::mgr)
            {
                bg::ShmDeleteSingleton<bg::detail::VptrManager>();
                bg::detail::mgr = nullptr;
            }
        }

        VptrManager::~VptrManager()
        {
            for(auto& item : mgr->m_type_to_containers)
            {
                item.second->~VptrObjectContainerBase();
                bg::ShmFree(item.second);
            }
            m_type_to_containers.clear();
        }

        void VptrManager::FixVptr()
        {
            for(auto& item : mgr->m_type_to_containers)
            {
                item.second->FixVptr();
            }
        }

        void* VptrManager::NewSpace(size_t bytes)
        {

            char* result = (char*)bg::ShmMalloc(bytes + 8);
            if(result)
            {
                result += 8;
            }
            return result;
        }

        void VptrManager::FreeSpace(void* ptr)
        {
            uintptr_t* raw_ptr = reinterpret_cast<uintptr_t*>(ptr);
            if(raw_ptr[0] == VPTR_MAGIC)
            {
                bg::ShmFree(raw_ptr);
            }
            else
            {
                SHM_ERROR("invalid magic(%#x) at address(%p).", raw_ptr[0], ptr);
            }
        }



        void* VptrManager::GetRealPtr(const TypeName& base_type, const TypeName& real_type, void* ptr)
        {
            if(base_type == real_type)
            {
                return ptr;
            }

            auto itor = bg::detail::mgr->m_type_to_containers.find(real_type);

            SHM_ASSERT_RETURN_VALUE(itor != bg::detail::mgr->m_type_to_containers.end(), ptr);

            auto info = itor->second->GetBaseTypeInfo(real_type);
            SHM_ASSERT_RETURN_VALUE(info, ptr);
            return &(((char*)ptr)[-info->offset]);
        }

        void VptrManager::RecordObject(const TypeName& type, void* ptr)
        {
            auto itor = bg::detail::mgr->m_type_to_containers.find(type);
            if(itor == bg::detail::mgr->m_type_to_containers.end())
            {

                VptrObjectContainerBase* base_ptr = hoc->NewShmContainer(type);
                SHM_ASSERT_RETURN_VOID(base_ptr && base_ptr->GetTypeName() == type);
                bg::detail::mgr->m_type_to_containers[type] = base_ptr;
                auto ret = bg::detail::mgr->m_type_to_containers.insert(std::make_pair(type, base_ptr));
                SHM_ASSERT_RETURN_VOID(ret.second && ret.first != mgr->m_type_to_containers.end());
                itor = ret.first;
            }


            itor->second->RecordObject(ptr);
            SHM_DEBUG("container(%p) for type(%s) created.", itor->second, type.c_str());
        }

        void VptrManager::RemoveObject(const TypeName& type, void* ptr)
        {
            auto itor = bg::detail::mgr->m_type_to_containers.find(type);

            SHM_ASSERT_RETURN_VOID(itor != bg::detail::mgr->m_type_to_containers.end());
            itor->second->RemoveObject(ptr);
        }


    }
}
